#  Copyright 2014 Cloudbase Solutions Srl
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import os
from unittest import mock

import ddt
from oslo_utils import timeutils
from oslo_utils import units

from cinder import context
from cinder import exception
from cinder.image import image_utils
from cinder.objects import fields
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.tests.unit import utils as test_utils
from cinder.volume.drivers import remotefs
from cinder.volume.drivers.windows import smbfs


@ddt.ddt
class WindowsSmbFsTestCase(test.TestCase):

    _FAKE_SHARE = '//1.2.3.4/share1'
    _FAKE_SHARE_HASH = 'db0bf952c1734092b83e8990bd321131'
    _FAKE_MNT_BASE = r'c:\openstack\mnt'
    _FAKE_MNT_POINT = os.path.join(_FAKE_MNT_BASE, _FAKE_SHARE_HASH)
    _FAKE_VOLUME_ID = '4f711859-4928-4cb7-801a-a50c37ceaccc'
    _FAKE_VOLUME_NAME = 'volume-%s.vhdx' % _FAKE_VOLUME_ID
    _FAKE_SNAPSHOT_ID = '50811859-4928-4cb7-801a-a50c37ceacba'
    _FAKE_SNAPSHOT_NAME = 'volume-%s-%s.vhdx' % (_FAKE_VOLUME_ID,
                                                 _FAKE_SNAPSHOT_ID)
    _FAKE_SNAPSHOT_PATH = os.path.join(_FAKE_MNT_POINT,
                                       _FAKE_SNAPSHOT_NAME)
    _FAKE_VOLUME_SIZE = 1
    _FAKE_TOTAL_SIZE = 2048
    _FAKE_TOTAL_AVAILABLE = 1024
    _FAKE_TOTAL_ALLOCATED = 1024
    _FAKE_SHARE_OPTS = '-o username=Administrator,password=12345'
    _FAKE_VOLUME_PATH = os.path.join(_FAKE_MNT_POINT,
                                     _FAKE_VOLUME_NAME)
    _FAKE_SHARE_OPTS = '-o username=Administrator,password=12345'

    @mock.patch.object(smbfs, 'utilsfactory')
    @mock.patch.object(smbfs, 'remotefs_brick')
    def setUp(self, mock_remotefs, mock_utilsfactory):
        super(WindowsSmbFsTestCase, self).setUp()

        self.context = context.get_admin_context()

        self._FAKE_SMBFS_CONFIG = mock.MagicMock(
            smbfs_shares_config=mock.sentinel.share_config_file,
            smbfs_default_volume_format='vhdx',
            nas_volume_prov_type='thin')

        self._smbfs_driver = smbfs.WindowsSmbfsDriver(
            configuration=self._FAKE_SMBFS_CONFIG)
        self._smbfs_driver._delete = mock.Mock()
        self._smbfs_driver._local_volume_dir = mock.Mock(
            return_value=self._FAKE_MNT_POINT)
        self._smbfs_driver.base = self._FAKE_MNT_BASE

        self._diskutils = self._smbfs_driver._diskutils
        self._vhdutils = self._smbfs_driver._vhdutils

        self.volume = self._simple_volume()
        self.snapshot = self._simple_snapshot(volume=self.volume)

        self._context = context.get_admin_context()
        self.updated_at = timeutils.utcnow()

    def _simple_volume(self, **kwargs):
        updates = {'id': self._FAKE_VOLUME_ID,
                   'size': self._FAKE_VOLUME_SIZE,
                   'provider_location': self._FAKE_SHARE}
        updates.update(kwargs)
        ctxt = context.get_admin_context()
        volume = test_utils.create_volume(ctxt, **updates)
        return volume

    def _simple_snapshot(self, **kwargs):
        volume = kwargs.pop('volume', None) or self._simple_volume()
        ctxt = context.get_admin_context()
        updates = {'id': self._FAKE_SNAPSHOT_ID,
                   'volume_id': volume.id}
        updates.update(kwargs)
        snapshot = test_utils.create_snapshot(ctxt, **updates)
        return snapshot

    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_check_os_platform')
    @mock.patch.object(remotefs.RemoteFSSnapDriverDistributed, 'do_setup')
    @mock.patch('os.path.exists')
    @mock.patch('os.path.isabs')
    @mock.patch.object(image_utils, 'check_qemu_img_version')
    def _test_setup(self, mock_check_qemu_img_version,
                    mock_is_abs, mock_exists,
                    mock_remotefs_do_setup,
                    mock_check_os_platform,
                    config, share_config_exists=True):
        mock_exists.return_value = share_config_exists
        fake_ensure_mounted = mock.MagicMock()
        self._smbfs_driver._ensure_shares_mounted = fake_ensure_mounted
        self._smbfs_driver._setup_pool_mappings = mock.Mock()
        self._smbfs_driver.configuration = config

        if not (config.smbfs_shares_config and share_config_exists):
            self.assertRaises(smbfs.SmbfsException,
                              self._smbfs_driver.do_setup,
                              mock.sentinel.context)
        else:
            self._smbfs_driver.do_setup(mock.sentinel.context)

            mock_check_qemu_img_version.assert_called_once_with(
                self._smbfs_driver._MINIMUM_QEMU_IMG_VERSION)
            mock_is_abs.assert_called_once_with(self._smbfs_driver.base)
            self.assertEqual({}, self._smbfs_driver.shares)
            fake_ensure_mounted.assert_called_once_with()
            self._smbfs_driver._setup_pool_mappings.assert_called_once_with()

        self.assertTrue(self._smbfs_driver._thin_provisioning_support)
        mock_check_os_platform.assert_called_once_with()

    def test_setup_pools(self):
        pool_mappings = {
            '//ip/share0': 'pool0',
            '//ip/share1': 'pool1',
        }
        self._smbfs_driver.configuration.smbfs_pool_mappings = pool_mappings
        self._smbfs_driver.shares = {
            '//ip/share0': None,
            '//ip/share1': None,
            '//ip/share2': None
        }

        expected_pool_mappings = pool_mappings.copy()
        expected_pool_mappings['//ip/share2'] = 'share2'

        self._smbfs_driver._setup_pool_mappings()
        self.assertEqual(expected_pool_mappings,
                         self._smbfs_driver._pool_mappings)

    def test_setup_pool_duplicates(self):
        self._smbfs_driver.configuration.smbfs_pool_mappings = {
            'share0': 'pool0',
            'share1': 'pool0'
        }
        self.assertRaises(smbfs.SmbfsException,
                          self._smbfs_driver._setup_pool_mappings)

    def test_initialize_connection(self):
        self._smbfs_driver.get_active_image_from_info = mock.Mock(
            return_value=self._FAKE_VOLUME_NAME)
        self._smbfs_driver._get_mount_point_base = mock.Mock(
            return_value=self._FAKE_MNT_BASE)
        self._smbfs_driver.shares = {self._FAKE_SHARE: self._FAKE_SHARE_OPTS}
        self._smbfs_driver.get_volume_format = mock.Mock(
            return_value=mock.sentinel.format)

        fake_data = {'export': self._FAKE_SHARE,
                     'format': mock.sentinel.format,
                     'name': self._FAKE_VOLUME_NAME,
                     'options': self._FAKE_SHARE_OPTS}
        expected = {
            'driver_volume_type': 'smbfs',
            'data': fake_data,
            'mount_point_base': self._FAKE_MNT_BASE}
        ret_val = self._smbfs_driver.initialize_connection(
            self.volume, None)

        self.assertEqual(expected, ret_val)

    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_get_snapshot_backing_file')
    @mock.patch.object(smbfs.WindowsSmbfsDriver, 'get_volume_format')
    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_get_mount_point_base')
    def test_initialize_connection_snapshot(self, mock_get_mount_base,
                                            mock_get_volume_format,
                                            mock_get_snap_by_backing_file):
        self._smbfs_driver.shares = {self._FAKE_SHARE: self._FAKE_SHARE_OPTS}
        mock_get_snap_by_backing_file.return_value = self._FAKE_VOLUME_NAME
        mock_get_volume_format.return_value = 'vhdx'
        mock_get_mount_base.return_value = self._FAKE_MNT_BASE

        exp_data = {'export': self._FAKE_SHARE,
                    'format': 'vhdx',
                    'name': self._FAKE_VOLUME_NAME,
                    'options': self._FAKE_SHARE_OPTS,
                    'access_mode': 'ro'}
        expected = {
            'driver_volume_type': 'smbfs',
            'data': exp_data,
            'mount_point_base': self._FAKE_MNT_BASE}
        ret_val = self._smbfs_driver.initialize_connection_snapshot(
            self.snapshot, mock.sentinel.connector)

        self.assertEqual(expected, ret_val)

        mock_get_snap_by_backing_file.assert_called_once_with(self.snapshot)
        mock_get_volume_format.assert_called_once_with(self.snapshot.volume)
        mock_get_mount_base.assert_called_once_with()

    def test_setup(self):
        self._test_setup(config=self._FAKE_SMBFS_CONFIG)

    def test_setup_missing_shares_config_option(self):
        fake_config = copy.copy(self._FAKE_SMBFS_CONFIG)
        fake_config.smbfs_shares_config = None
        self._test_setup(config=fake_config,
                         share_config_exists=False)

    def test_setup_missing_shares_config_file(self):
        self._test_setup(config=self._FAKE_SMBFS_CONFIG,
                         share_config_exists=False)

    @mock.patch.object(smbfs, 'context')
    @mock.patch.object(smbfs.WindowsSmbfsDriver,
                       '_get_pool_name_from_share')
    def test_get_total_allocated(self, mock_get_pool_name, mock_ctxt):
        fake_pool_name = 'pool0'
        fake_host_name = 'fake_host@fake_backend'
        fake_vol_sz_sum = 5

        mock_db = mock.Mock()
        mock_db.volume_data_get_for_host.return_value = [
            mock.sentinel.vol_count, fake_vol_sz_sum]

        self._smbfs_driver.host = fake_host_name
        self._smbfs_driver.db = mock_db

        mock_get_pool_name.return_value = fake_pool_name

        allocated = self._smbfs_driver._get_total_allocated(
            mock.sentinel.share)
        self.assertEqual(fake_vol_sz_sum << 30,
                         allocated)

        mock_get_pool_name.assert_called_once_with(mock.sentinel.share)
        mock_db.volume_data_get_for_host.assert_called_once_with(
            context=mock_ctxt.get_admin_context.return_value,
            host='fake_host@fake_backend#pool0')

    @mock.patch.object(smbfs.WindowsSmbfsDriver,
                       '_get_local_volume_path_template')
    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_lookup_local_volume_path')
    @mock.patch.object(smbfs.WindowsSmbfsDriver, 'get_volume_format')
    def _test_get_volume_path(self, mock_get_volume_format, mock_lookup_volume,
                              mock_get_path_template, volume_exists=True):
        drv = self._smbfs_driver
        (mock_get_path_template.return_value,
         ext) = os.path.splitext(self._FAKE_VOLUME_PATH)
        volume_format = ext.strip('.')

        mock_lookup_volume.return_value = (
            self._FAKE_VOLUME_PATH if volume_exists else None)
        mock_get_volume_format.return_value = volume_format

        ret_val = drv.local_path(self.volume)

        if volume_exists:
            self.assertFalse(mock_get_volume_format.called)
        else:
            mock_get_volume_format.assert_called_once_with(self.volume)
        self.assertEqual(self._FAKE_VOLUME_PATH, ret_val)

    def test_get_existing_volume_path(self):
        self._test_get_volume_path()

    def test_get_new_volume_path(self):
        self._test_get_volume_path(volume_exists=False)

    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_local_volume_dir')
    def test_get_local_volume_path_template(self, mock_get_local_dir):
        mock_get_local_dir.return_value = self._FAKE_MNT_POINT
        ret_val = self._smbfs_driver._get_local_volume_path_template(
            self.volume)
        exp_template = os.path.splitext(self._FAKE_VOLUME_PATH)[0]
        self.assertEqual(exp_template, ret_val)

    @mock.patch('os.path.exists')
    def test_lookup_local_volume_path(self, mock_exists):
        expected_path = self._FAKE_VOLUME_PATH + '.vhdx'
        mock_exists.side_effect = lambda x: x == expected_path

        ret_val = self._smbfs_driver._lookup_local_volume_path(
            self._FAKE_VOLUME_PATH)

        extensions = [
            ".%s" % ext
            for ext in self._smbfs_driver._VALID_IMAGE_EXTENSIONS]
        possible_paths = [self._FAKE_VOLUME_PATH + ext
                          for ext in extensions]
        mock_exists.assert_has_calls(
            [mock.call(path) for path in possible_paths])
        self.assertEqual(expected_path, ret_val)

    @mock.patch.object(smbfs.WindowsSmbfsDriver,
                       '_get_local_volume_path_template')
    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_lookup_local_volume_path')
    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_get_volume_format_spec')
    def _test_get_volume_format(self, mock_get_format_spec,
                                mock_lookup_volume, mock_get_path_template,
                                qemu_format=False, volume_format='vhdx',
                                expected_vol_fmt=None,
                                volume_exists=True):
        expected_vol_fmt = expected_vol_fmt or volume_format

        vol_path = '%s.%s' % (os.path.splitext(self._FAKE_VOLUME_PATH)[0],
                              volume_format)
        mock_get_path_template.return_value = vol_path
        mock_lookup_volume.return_value = (
            vol_path if volume_exists else None)

        mock_get_format_spec.return_value = volume_format

        supported_fmts = self._smbfs_driver._SUPPORTED_IMAGE_FORMATS
        if volume_format.lower() not in supported_fmts:
            self.assertRaises(smbfs.SmbfsException,
                              self._smbfs_driver.get_volume_format,
                              self.volume,
                              qemu_format)

        else:
            ret_val = self._smbfs_driver.get_volume_format(self.volume,
                                                           qemu_format)

            if volume_exists:
                self.assertFalse(mock_get_format_spec.called)
            else:
                mock_get_format_spec.assert_called_once_with(self.volume)

            self.assertEqual(expected_vol_fmt, ret_val)

    def test_get_volume_format_invalid_extension(self):
        self._test_get_volume_format(volume_format='fake')

    def test_get_existing_vhdx_volume_format(self):
        self._test_get_volume_format()

    def test_get_new_vhd_volume_format(self):
        fmt = 'vhd'
        self._test_get_volume_format(volume_format=fmt,
                                     volume_exists=False,
                                     expected_vol_fmt=fmt)

    def test_get_new_vhd_legacy_volume_format(self):
        img_fmt = 'vhd'
        expected_fmt = 'vpc'
        self._test_get_volume_format(volume_format=img_fmt,
                                     volume_exists=False,
                                     qemu_format=True,
                                     expected_vol_fmt=expected_fmt)

    @ddt.data([False, False],
              [True, True],
              [False, True])
    @ddt.unpack
    def test_get_volume_format_spec(self,
                                    volume_meta_contains_fmt,
                                    volume_type_contains_fmt):
        self._smbfs_driver.configuration = copy.copy(self._FAKE_SMBFS_CONFIG)

        fake_vol_meta_fmt = 'vhd'
        fake_vol_type_fmt = 'vhdx'

        volume_metadata = {}
        volume_type_extra_specs = {}

        if volume_meta_contains_fmt:
            volume_metadata['volume_format'] = fake_vol_meta_fmt
        elif volume_type_contains_fmt:
            volume_type_extra_specs['smbfs:volume_format'] = fake_vol_type_fmt

        volume_type = fake_volume.fake_volume_type_obj(self.context)
        volume = fake_volume.fake_volume_obj(self.context)
        # Optional arguments are not set in _from_db_object,
        # so have to set explicitly here
        volume.volume_type = volume_type
        volume.metadata = volume_metadata
        # Same for extra_specs and VolumeType
        volume_type.extra_specs = volume_type_extra_specs

        resulted_fmt = self._smbfs_driver._get_volume_format_spec(volume)

        if volume_meta_contains_fmt:
            expected_fmt = fake_vol_meta_fmt
        elif volume_type_contains_fmt:
            expected_fmt = fake_vol_type_fmt
        else:
            expected_fmt = self._FAKE_SMBFS_CONFIG.smbfs_default_volume_format

        self.assertEqual(expected_fmt, resulted_fmt)

    @mock.patch.object(remotefs.RemoteFSSnapDriverDistributed,
                       'create_volume')
    def test_create_volume_base(self, mock_create_volume):
        self._smbfs_driver.create_volume(self.volume)
        mock_create_volume.assert_called_once_with(self.volume)

    @mock.patch('os.path.exists')
    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_get_vhd_type')
    def _test_create_volume(self, mock_get_vhd_type, mock_exists,
                            volume_exists=False, volume_format='vhdx'):
        mock_exists.return_value = volume_exists
        self._smbfs_driver.create_vhd = mock.MagicMock()
        fake_create = self._smbfs_driver._vhdutils.create_vhd
        self._smbfs_driver.get_volume_format = mock.Mock(
            return_value=volume_format)

        if volume_exists or volume_format not in ('vhd', 'vhdx'):
            self.assertRaises(exception.InvalidVolume,
                              self._smbfs_driver._do_create_volume,
                              self.volume)
        else:
            fake_vol_path = self._FAKE_VOLUME_PATH
            self._smbfs_driver._do_create_volume(self.volume)
            fake_create.assert_called_once_with(
                fake_vol_path, mock_get_vhd_type.return_value,
                max_internal_size=self.volume.size << 30,
                guid=self.volume.id)

    def test_create_volume(self):
        self._test_create_volume()

    def test_create_existing_volume(self):
        self._test_create_volume(True)

    def test_create_volume_invalid_volume(self):
        self._test_create_volume(volume_format="qcow")

    def test_delete_volume(self):
        drv = self._smbfs_driver
        fake_vol_info = self._FAKE_VOLUME_PATH + '.info'

        drv._ensure_share_mounted = mock.MagicMock()
        fake_ensure_mounted = drv._ensure_share_mounted

        drv._local_volume_dir = mock.Mock(
            return_value=self._FAKE_MNT_POINT)
        drv.get_active_image_from_info = mock.Mock(
            return_value=self._FAKE_VOLUME_NAME)
        drv._delete = mock.Mock()
        drv._local_path_volume_info = mock.Mock(
            return_value=fake_vol_info)

        with mock.patch('os.path.exists', lambda x: True):
            drv.delete_volume(self.volume)

            fake_ensure_mounted.assert_called_once_with(self._FAKE_SHARE)
            drv._delete.assert_any_call(
                self._FAKE_VOLUME_PATH)
            drv._delete.assert_any_call(fake_vol_info)

    def test_ensure_mounted(self):
        self._smbfs_driver.shares = {self._FAKE_SHARE: self._FAKE_SHARE_OPTS}

        self._smbfs_driver._ensure_share_mounted(self._FAKE_SHARE)
        self._smbfs_driver._remotefsclient.mount.assert_called_once_with(
            self._FAKE_SHARE, self._FAKE_SHARE_OPTS)

    def test_get_capacity_info(self):
        self._diskutils.get_disk_capacity.return_value = (
            self._FAKE_TOTAL_SIZE, self._FAKE_TOTAL_AVAILABLE)
        self._smbfs_driver._get_mount_point_for_share = mock.Mock(
            return_value=mock.sentinel.mnt_point)
        self._smbfs_driver._get_total_allocated = mock.Mock(
            return_value=self._FAKE_TOTAL_ALLOCATED)

        ret_val = self._smbfs_driver._get_capacity_info(self._FAKE_SHARE)
        expected_ret_val = [int(x) for x in [self._FAKE_TOTAL_SIZE,
                            self._FAKE_TOTAL_AVAILABLE,
                            self._FAKE_TOTAL_ALLOCATED]]
        self.assertEqual(expected_ret_val, ret_val)

        self._smbfs_driver._get_mount_point_for_share.assert_called_once_with(
            self._FAKE_SHARE)
        self._diskutils.get_disk_capacity.assert_called_once_with(
            mock.sentinel.mnt_point)
        self._smbfs_driver._get_total_allocated.assert_called_once_with(
            self._FAKE_SHARE)

    def _test_get_img_info(self, backing_file=None):
        self._smbfs_driver._vhdutils.get_vhd_parent_path.return_value = (
            backing_file)

        image_info = self._smbfs_driver._qemu_img_info(self._FAKE_VOLUME_PATH)
        self.assertEqual(self._FAKE_VOLUME_NAME,
                         image_info.image)
        backing_file_name = backing_file and os.path.basename(backing_file)
        self.assertEqual(backing_file_name, image_info.backing_file)

    def test_get_img_info_without_backing_file(self):
        self._test_get_img_info()

    def test_get_snapshot_info(self):
        self._test_get_img_info(self._FAKE_VOLUME_PATH)

    @ddt.data('attached', 'detached')
    def test_create_snapshot(self, attach_status):
        self.snapshot.volume.attach_status = attach_status
        self.snapshot.volume.save()

        self._smbfs_driver._vhdutils.create_differencing_vhd = (
            mock.Mock())
        self._smbfs_driver._local_volume_dir = mock.Mock(
            return_value=self._FAKE_MNT_POINT)

        fake_create_diff = (
            self._smbfs_driver._vhdutils.create_differencing_vhd)

        self._smbfs_driver._do_create_snapshot(
            self.snapshot,
            os.path.basename(self._FAKE_VOLUME_PATH),
            self._FAKE_SNAPSHOT_PATH)

        if attach_status != 'attached':
            fake_create_diff.assert_called_once_with(self._FAKE_SNAPSHOT_PATH,
                                                     self._FAKE_VOLUME_PATH)
        else:
            fake_create_diff.assert_not_called()

        self.assertEqual(os.path.basename(self._FAKE_VOLUME_PATH),
                         self.snapshot.metadata['backing_file'])
        # Ensure that the changes have been saved.
        self.assertFalse(bool(self.snapshot.obj_what_changed()))

    @mock.patch.object(smbfs.WindowsSmbfsDriver,
                       '_check_extend_volume_support')
    @mock.patch.object(smbfs.WindowsSmbfsDriver,
                       '_local_path_active_image')
    def test_extend_volume(self, mock_get_active_img,
                           mock_check_ext_support):
        volume = fake_volume.fake_volume_obj(self.context)
        new_size = volume.size + 1

        self._smbfs_driver.extend_volume(volume, new_size)

        mock_check_ext_support.assert_called_once_with(volume, new_size)
        mock_get_active_img.assert_called_once_with(volume)
        self._vhdutils.resize_vhd.assert_called_once_with(
            mock_get_active_img.return_value,
            new_size * units.Gi,
            is_file_max_size=False)

    @ddt.data({'snapshots_exist': True},
              {'vol_fmt': smbfs.WindowsSmbfsDriver._DISK_FORMAT_VHD,
               'snapshots_exist': True,
               'expected_exc': exception.InvalidVolume})
    @ddt.unpack
    @mock.patch.object(smbfs.WindowsSmbfsDriver,
                       'get_volume_format')
    @mock.patch.object(smbfs.WindowsSmbfsDriver,
                       '_snapshots_exist')
    def test_check_extend_support(self, mock_snapshots_exist,
                                  mock_get_volume_format,
                                  vol_fmt=None, snapshots_exist=False,
                                  share_eligible=True,
                                  expected_exc=None):
        vol_fmt = vol_fmt or self._smbfs_driver._DISK_FORMAT_VHDX

        volume = fake_volume.fake_volume_obj(
            self.context, provider_location='fake_provider_location')
        new_size = volume.size + 1

        mock_snapshots_exist.return_value = snapshots_exist
        mock_get_volume_format.return_value = vol_fmt

        if expected_exc:
            self.assertRaises(expected_exc,
                              self._smbfs_driver._check_extend_volume_support,
                              volume, new_size)
        else:
            self._smbfs_driver._check_extend_volume_support(volume, new_size)

            mock_get_volume_format.assert_called_once_with(volume)
            mock_snapshots_exist.assert_called_once_with(volume)

    @ddt.data({},
              {'delete_latest': True},
              {'attach_status': 'detached'},
              {'snap_info_contains_snap_id': False})
    @ddt.unpack
    @mock.patch.object(remotefs.RemoteFSSnapDriverDistributed,
                       '_delete_snapshot')
    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_local_volume_dir')
    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_local_path_volume_info')
    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_write_info_file')
    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_read_info_file')
    @mock.patch.object(smbfs.WindowsSmbfsDriver,
                       '_nova_assisted_vol_snap_delete')
    @mock.patch.object(smbfs.WindowsSmbfsDriver,
                       '_get_snapshot_by_backing_file')
    def test_delete_snapshot(self, mock_get_snap_by_backing_file,
                             mock_nova_assisted_snap_del,
                             mock_read_info_file, mock_write_info_file,
                             mock_local_path_volume_info,
                             mock_get_local_dir,
                             mock_remotefs_snap_delete,
                             attach_status='attached',
                             snap_info_contains_snap_id=True,
                             delete_latest=False):
        self.snapshot.volume.attach_status = attach_status
        self.snapshot.metadata['backing_file'] = os.path.basename(
            self._FAKE_VOLUME_PATH)
        higher_snapshot = self._simple_snapshot(id=None,
                                                volume=self.volume)

        fake_snap_file = 'snap_file'
        fake_snap_parent_path = os.path.join(self._FAKE_MNT_POINT,
                                             'snap_file_parent')
        active_img = 'active_img' if not delete_latest else fake_snap_file

        snap_info = dict(active=active_img)
        if snap_info_contains_snap_id:
            snap_info[self.snapshot.id] = fake_snap_file

        mock_get_snap_by_backing_file.return_value = (
            higher_snapshot if not delete_latest else None)
        mock_info_path = mock_local_path_volume_info.return_value
        mock_read_info_file.return_value = snap_info
        mock_get_local_dir.return_value = self._FAKE_MNT_POINT
        self._vhdutils.get_vhd_parent_path.return_value = (
            fake_snap_parent_path)

        expected_delete_info = {'file_to_merge': fake_snap_file,
                                'volume_id': self.snapshot.volume.id}

        self._smbfs_driver._delete_snapshot(self.snapshot)

        if attach_status != 'attached':
            mock_remotefs_snap_delete.assert_called_once_with(self.snapshot)
        elif snap_info_contains_snap_id:
            mock_local_path_volume_info.assert_called_once_with(
                self.snapshot.volume)
            mock_read_info_file.assert_called_once_with(
                mock_info_path, empty_if_missing=True)
            mock_nova_assisted_snap_del.assert_called_once_with(
                self.snapshot._context, self.snapshot, expected_delete_info)

            exp_merged_img_path = os.path.join(self._FAKE_MNT_POINT,
                                               fake_snap_file)
            self._smbfs_driver._delete.assert_called_once_with(
                exp_merged_img_path)

            if delete_latest:
                self._vhdutils.get_vhd_parent_path.assert_called_once_with(
                    exp_merged_img_path)
                exp_active = os.path.basename(fake_snap_parent_path)
            else:
                exp_active = active_img

            self.assertEqual(exp_active, snap_info['active'])
            self.assertNotIn(snap_info, self.snapshot.id)
            mock_write_info_file.assert_called_once_with(mock_info_path,
                                                         snap_info)

        if attach_status != 'attached' or not snap_info_contains_snap_id:
            mock_nova_assisted_snap_del.assert_not_called()
            mock_write_info_file.assert_not_called()

        if not delete_latest and snap_info_contains_snap_id:
            self.assertEqual(os.path.basename(self._FAKE_VOLUME_PATH),
                             higher_snapshot.metadata['backing_file'])
            self.assertFalse(bool(higher_snapshot.obj_what_changed()))

    @ddt.data(True, False)
    def test_get_snapshot_by_backing_file(self, metadata_set):
        backing_file = 'fake_backing_file'
        if metadata_set:
            self.snapshot.metadata['backing_file'] = backing_file
            self.snapshot.save()

        for idx in range(2):
            # We're adding a few other snapshots.
            self._simple_snapshot(id=None,
                                  volume=self.volume)

        snapshot = self._smbfs_driver._get_snapshot_by_backing_file(
            self.volume, backing_file)

        if metadata_set:
            self.assertEqual(self.snapshot.id, snapshot.id)
        else:
            self.assertIsNone(snapshot)

    @ddt.data(True, False)
    @mock.patch.object(remotefs.RemoteFSSnapDriverDistributed,
                       '_get_snapshot_backing_file')
    def test_get_snapshot_backing_file_md_set(self, md_set,
                                              remotefs_get_backing_file):
        backing_file = 'fake_backing_file'
        if md_set:
            self.snapshot.metadata['backing_file'] = backing_file

        ret_val = self._smbfs_driver._get_snapshot_backing_file(
            self.snapshot)

        # If the metadata is not set, we expect the super class method to
        # be used, which is supposed to query the image.
        if md_set:
            self.assertEqual(backing_file, ret_val)
        else:
            self.assertEqual(remotefs_get_backing_file.return_value,
                             ret_val)
            remotefs_get_backing_file.assert_called_once_with(
                self.snapshot)

    def test_create_volume_from_unavailable_snapshot(self):
        self.snapshot.status = fields.SnapshotStatus.ERROR
        self.assertRaises(
            exception.InvalidSnapshot,
            self._smbfs_driver.create_volume_from_snapshot,
            self.volume, self.snapshot)

    @ddt.data(True, False)
    def test_copy_volume_to_image(self, has_parent=False):
        drv = self._smbfs_driver

        volume = test_utils.create_volume(
            self._context, volume_type_id=fake.VOLUME_TYPE_ID,
            updated_at=self.updated_at)

        extra_specs = {
            'image_service:store_id': 'fake-store'
        }
        test_utils.create_volume_type(self._context.elevated(),
                                      id=fake.VOLUME_TYPE_ID, name="test_type",
                                      extra_specs=extra_specs)

        fake_image_meta = {'id': 'fake-image-id'}
        fake_img_format = self._smbfs_driver._DISK_FORMAT_VHDX

        if has_parent:
            fake_volume_path = self._FAKE_SNAPSHOT_PATH
            fake_parent_path = self._FAKE_VOLUME_PATH
        else:
            fake_volume_path = self._FAKE_VOLUME_PATH
            fake_parent_path = None

        fake_active_image = os.path.basename(fake_volume_path)

        drv.get_active_image_from_info = mock.Mock(
            return_value=fake_active_image)
        drv._local_volume_dir = mock.Mock(
            return_value=self._FAKE_MNT_POINT)
        drv.get_volume_format = mock.Mock(
            return_value=fake_img_format)
        drv._vhdutils.get_vhd_parent_path.return_value = (
            fake_parent_path)

        with mock.patch.object(image_utils, 'upload_volume') as (
                fake_upload_volume):
            drv.copy_volume_to_image(
                mock.sentinel.context, volume,
                mock.sentinel.image_service, fake_image_meta)

            if has_parent:
                fake_temp_image_name = '%s.temp_image.%s.%s' % (
                    volume.id,
                    fake_image_meta['id'],
                    fake_img_format)
                fake_temp_image_path = os.path.join(
                    self._FAKE_MNT_POINT,
                    fake_temp_image_name)
                fake_active_image_path = os.path.join(
                    self._FAKE_MNT_POINT,
                    fake_active_image)
                upload_path = fake_temp_image_path

                drv._vhdutils.convert_vhd.assert_called_once_with(
                    fake_active_image_path,
                    fake_temp_image_path)
                drv._delete.assert_called_once_with(
                    fake_temp_image_path)
            else:
                upload_path = fake_volume_path

            fake_upload_volume.assert_called_once_with(
                mock.sentinel.context, mock.sentinel.image_service,
                fake_image_meta, upload_path, volume_format=fake_img_format,
                store_id='fake-store', base_image_ref=None, compress=True,
                run_as_root=True)

    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_get_vhd_type')
    def test_copy_image_to_volume(self, mock_get_vhd_type):
        drv = self._smbfs_driver

        drv.get_volume_format = mock.Mock(
            return_value=mock.sentinel.volume_format)
        drv.local_path = mock.Mock(
            return_value=self._FAKE_VOLUME_PATH)
        drv.configuration = mock.MagicMock()
        drv.configuration.volume_dd_blocksize = mock.sentinel.block_size

        with mock.patch.object(image_utils,
                               'fetch_to_volume_format') as fake_fetch:
            drv.copy_image_to_volume(
                mock.sentinel.context, self.volume,
                mock.sentinel.image_service,
                mock.sentinel.image_id)
            fake_fetch.assert_called_once_with(
                mock.sentinel.context,
                mock.sentinel.image_service,
                mock.sentinel.image_id,
                self._FAKE_VOLUME_PATH, mock.sentinel.volume_format,
                mock.sentinel.block_size,
                mock_get_vhd_type.return_value)
            drv._vhdutils.resize_vhd.assert_called_once_with(
                self._FAKE_VOLUME_PATH,
                self.volume.size * units.Gi,
                is_file_max_size=False)
            drv._vhdutils.set_vhd_guid.assert_called_once_with(
                self._FAKE_VOLUME_PATH,
                self.volume.id)

    @mock.patch.object(smbfs.WindowsSmbfsDriver, '_get_vhd_type')
    def test_copy_volume_from_snapshot(self, mock_get_vhd_type):
        drv = self._smbfs_driver

        drv._get_snapshot_backing_file = mock.Mock(
            return_value=self._FAKE_VOLUME_NAME)
        drv._local_volume_dir = mock.Mock(
            return_value=self._FAKE_MNT_POINT)
        drv.local_path = mock.Mock(
            return_value=mock.sentinel.new_volume_path)

        drv._copy_volume_from_snapshot(self.snapshot,
                                       self.volume, self.volume.size)

        drv._get_snapshot_backing_file.assert_called_once_with(
            self.snapshot)
        drv._delete.assert_called_once_with(mock.sentinel.new_volume_path)
        drv._vhdutils.convert_vhd.assert_called_once_with(
            self._FAKE_VOLUME_PATH,
            mock.sentinel.new_volume_path,
            vhd_type=mock_get_vhd_type.return_value)
        drv._vhdutils.set_vhd_guid.assert_called_once_with(
            mock.sentinel.new_volume_path,
            self.volume.id)
        drv._vhdutils.resize_vhd.assert_called_once_with(
            mock.sentinel.new_volume_path,
            self.volume.size * units.Gi,
            is_file_max_size=False)

    def test_copy_encrypted_volume_from_snapshot(self):
        # We expect an exception to be raised if an encryption
        # key is provided since we don't support encryted volumes
        # for the time being.
        self.assertRaises(exception.NotSupportedOperation,
                          self._smbfs_driver._copy_volume_from_snapshot,
                          self.snapshot, self.volume,
                          self.volume.size,
                          mock.sentinel.src_key,
                          mock.sentinel.dest_key)

    def test_rebase_img(self):
        drv = self._smbfs_driver
        drv._rebase_img(
            self._FAKE_SNAPSHOT_PATH,
            self._FAKE_VOLUME_NAME, 'vhdx')
        drv._vhdutils.reconnect_parent_vhd.assert_called_once_with(
            self._FAKE_SNAPSHOT_PATH, self._FAKE_VOLUME_PATH)

    def test_copy_volume_image(self):
        self._smbfs_driver._copy_volume_image(mock.sentinel.src,
                                              mock.sentinel.dest)
        self._smbfs_driver._pathutils.copy.assert_called_once_with(
            mock.sentinel.src, mock.sentinel.dest)

    def test_get_pool_name_from_share(self):
        self._smbfs_driver._pool_mappings = {
            mock.sentinel.share: mock.sentinel.pool}

        pool = self._smbfs_driver._get_pool_name_from_share(
            mock.sentinel.share)
        self.assertEqual(mock.sentinel.pool, pool)

    def test_get_share_from_pool_name(self):
        self._smbfs_driver._pool_mappings = {
            mock.sentinel.share: mock.sentinel.pool}

        share = self._smbfs_driver._get_share_from_pool_name(
            mock.sentinel.pool)
        self.assertEqual(mock.sentinel.share, share)

    def test_get_pool_name_from_share_exception(self):
        self._smbfs_driver._pool_mappings = {}

        self.assertRaises(smbfs.SmbfsException,
                          self._smbfs_driver._get_share_from_pool_name,
                          mock.sentinel.pool)

    def test_get_vhd_type(self):
        drv = self._smbfs_driver

        mock_type = drv._get_vhd_type(qemu_subformat=True)
        self.assertEqual(mock_type, 'dynamic')

        mock_type = drv._get_vhd_type(qemu_subformat=False)
        self.assertEqual(mock_type, 3)

        self._smbfs_driver.configuration.nas_volume_prov_type = (
            'thick')

        mock_type = drv._get_vhd_type(qemu_subformat=True)
        self.assertEqual(mock_type, 'fixed')

    def test_get_managed_vol_expected_path(self):
        self._vhdutils.get_vhd_format.return_value = 'vhdx'
        vol_location = dict(vol_local_path=mock.sentinel.image_path,
                            mountpoint=self._FAKE_MNT_POINT)

        path = self._smbfs_driver._get_managed_vol_expected_path(
            self.volume, vol_location)
        self.assertEqual(self._FAKE_VOLUME_PATH, path)

        self._vhdutils.get_vhd_format.assert_called_once_with(
            mock.sentinel.image_path)

    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       'manage_existing')
    def test_manage_existing(self, remotefs_manage):
        model_update = dict(provider_location=self._FAKE_SHARE)
        remotefs_manage.return_value = model_update
        self._smbfs_driver.local_path = mock.Mock(
            return_value=mock.sentinel.vol_path)

        # Let's make sure that the provider location gets set.
        # It's needed by self.local_path.
        self.volume.provider_location = None

        ret_val = self._smbfs_driver.manage_existing(
            self.volume, mock.sentinel.ref)

        self.assertEqual(model_update, ret_val)
        self.assertEqual(self._FAKE_SHARE, self.volume.provider_location)

        self._vhdutils.set_vhd_guid.assert_called_once_with(
            mock.sentinel.vol_path,
            self.volume.id)
        self._smbfs_driver.local_path.assert_called_once_with(self.volume)
        remotefs_manage.assert_called_once_with(self.volume, mock.sentinel.ref)



#    (c) Copyright 2019 Brocade, a Broadcom Company
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Common constants used by Brocade FC Zone Driver.
"""
YES = 'y'
ACTIVE_ZONE_CONFIG = 'active_zone_config'
CFG_ZONESET = 'cfg:'
CFG_ZONES = 'zones'
OPENSTACK_CFG_NAME = 'OpenStack_Cfg'
SUCCESS = 'Success'
TRANS_ABORTABLE = 'It is abortable'

"""
CLI Commands for FC zoning operations.
"""
GET_ACTIVE_ZONE_CFG = 'cfgactvshow'
ZONE_CREATE = 'zonecreate '
ZONESET_CREATE = 'cfgcreate '
CFG_SAVE = 'cfgsave'
CFG_ADD = 'cfgadd '
ACTIVATE_ZONESET = 'cfgenable '
DEACTIVATE_ZONESET = 'cfgdisable'
CFG_DELETE = 'cfgdelete '
CFG_REMOVE = 'cfgremove '
ZONE_DELETE = 'zonedelete '
ZONE_ADD = 'zoneadd '
ZONE_REMOVE = 'zoneremove '
CFG_SHOW_TRANS = 'cfgtransshow'
CFG_ZONE_TRANS_ABORT = 'cfgtransabort'
NS_SHOW = 'nsshow'
NS_CAM_SHOW = 'nscamshow'

"""
HTTPS connector constants
"""
AUTH_HEADER = "Authorization"
PROTOCOL_HTTPS = "HTTPS"
STATUS_OK = 200
SECINFO_PAGE = "/secinfo.html"
AUTHEN_PAGE = "/authenticate.html"
GET_METHOD = "GET"
POST_METHOD = "POST"
SECINFO_BEGIN = "--BEGIN SECINFO"
SECINFO_END = "--END SECINFO"
RANDOM = "RANDOM"
AUTHVERSION = "AUTHVERSION"
IDENTIFIER = "Identifier"
AUTH_STRING = "Custom_Basic "  # Trailing space is required, do not remove
AUTHEN_BEGIN = "--BEGIN AUTHENTICATE"
AUTHEN_END = "--END AUTHENTICATE"
AUTHENTICATED = "authenticated"
SESSION_PAGE_ACTION = "/session.html?action=query"
SESSION_BEGIN = "--BEGIN SESSION"
SESSION_END = "--END SESSION"
SESSION_PAGE = "/session.html"
LOGOUT_PAGE = "/logout.html"
ZONEINFO_BEGIN = "--BEGIN ZONE INFO"
ZONEINFO_END = "--END ZONE INFO"
SWITCH_PAGE = "/switch.html"
SWITCHINFO_BEGIN = "--BEGIN SWITCH INFORMATION"
SWITCHINFO_END = "--END SWITCH INFORMATION"
FIRMWARE_VERSION = "swFWVersion"
VF_ENABLED = "vfEnabled"
MANAGEABLE_VF = "manageableLFList"
CHANGE_VF = ("Session=--BEGIN SESSION\n\taction=apply\n\tLFId=  {vfid}  "
             "\b\t--END SESSION")
ZONE_TRAN_STATUS = "/gzoneinfo.htm?txnId={txnId}"
CFG_DELIM = "\x01"
ZONE_DELIM = "\x02"
ALIAS_DELIM = "\x03"
QLP_DELIM = "\x04"
ZONE_END_DELIM = "\x05&saveonly="
IFA_DELIM = "\x06"
ACTIVE_CFG_DELIM = "\x07"
DEFAULT_CFG = "d__efault__Cfg"
NS_PAGE = "/nsinfo.htm?format=1&type=all"
NS_DELIM = "deviceport.portwwn="
ZONE_TX_BEGIN = "--BEGIN ZONE_TXN_INFO"
ZONE_TX_END = "--END ZONE_TXN_INFO"
ZONE_ERROR_CODE = "errorCode"
ZONE_PAGE = "/gzoneinfo.htm"
CFG_NAME = "openstack_cfg"
ZONE_STRING_PREFIX = "zonecfginfo="
ZONE_ERROR_MSG = "errorMessage"
ZONE_TX_ID = "txnId"
ZONE_TX_STATUS = "status"
SESSION_LF_ID = "sessionLFId"
HTTP = "http"
HTTPS = "https"
REST_HTTP = "rest_http"
REST_HTTPS = "rest_https"



# Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for volume name_id."""

from oslo_config import cfg

from cinder import context
from cinder import db
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder.tests.unit import utils as testutils


CONF = cfg.CONF


class NameIDsTestCase(test.TestCase):
    """Test cases for naming volumes with name_id."""

    def setUp(self):
        super(NameIDsTestCase, self).setUp()
        self.ctxt = context.RequestContext(user_id=fake.USER_ID,
                                           project_id=fake.PROJECT_ID)

    def test_name_id_same(self):
        """New volume should have same 'id' and 'name_id'."""
        vol_ref = testutils.create_volume(self.ctxt, size=1)
        self.assertEqual(vol_ref['name_id'], vol_ref['id'])
        expected_name = CONF.volume_name_template % vol_ref['id']
        self.assertEqual(expected_name, vol_ref['name'])

    def test_name_id_diff(self):
        """Change name ID to mimic volume after migration."""
        vol_ref = testutils.create_volume(self.ctxt, size=1,
                                          _name_id=fake.VOLUME2_ID)
        vol_ref = db.volume_get(self.ctxt, vol_ref['id'])
        expected_name = CONF.volume_name_template % fake.VOLUME2_ID
        self.assertEqual(expected_name, vol_ref['name'])

    def test_name_id_snapshot_volume_name(self):
        """Make sure snapshot['volume_name'] is updated."""
        vol_ref = testutils.create_volume(self.ctxt, size=1,
                                          _name_id=fake.VOLUME2_ID)
        snap_ref = testutils.create_snapshot(self.ctxt, vol_ref['id'])
        expected_name = CONF.volume_name_template % fake.VOLUME2_ID
        self.assertEqual(expected_name, snap_ref['volume_name'])



# Copyright (c) 2010 OpenStack Foundation
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Scheduler base class that all Schedulers should inherit from
"""

from oslo_config import cfg
from oslo_utils import importutils
from oslo_utils import timeutils

from cinder.i18n import _
from cinder import objects
from cinder.volume import rpcapi as volume_rpcapi


scheduler_driver_opts = [
    cfg.StrOpt('scheduler_host_manager',
               default='cinder.scheduler.host_manager.HostManager',
               help='The scheduler host manager class to use'),
    cfg.IntOpt('scheduler_max_attempts',
               default=3,
               help='Maximum number of attempts to schedule a volume'),
]

CONF = cfg.CONF
CONF.register_opts(scheduler_driver_opts)


def volume_update_db(context, volume_id, host, cluster_name,
                     availability_zone=None):
    """Set the host, cluster_name, and set the scheduled_at field of a volume.

    :returns: A Volume with the updated fields set properly.
    """
    volume = objects.Volume.get_by_id(context, volume_id)
    volume.host = host
    volume.cluster_name = cluster_name
    volume.scheduled_at = timeutils.utcnow()
    volume.availability_zone = availability_zone
    volume.save()

    # A volume object is expected to be returned, as it is used by
    # filter_scheduler.
    return volume


def generic_group_update_db(context, group, host, cluster_name):
    """Set the host and the scheduled_at field of a group.

    :returns: A Group with the updated fields set properly.
    """
    group.update({'host': host, 'updated_at': timeutils.utcnow(),
                  'cluster_name': cluster_name})
    group.save()
    return group


class Scheduler(object):
    """The base class that all Scheduler classes should inherit from."""

    def __init__(self):
        self.host_manager = importutils.import_object(
            CONF.scheduler_host_manager)
        self.volume_rpcapi = volume_rpcapi.VolumeAPI()

    def reset(self):
        """Reset volume RPC API object to load new version pins."""
        self.volume_rpcapi = volume_rpcapi.VolumeAPI()

    def is_ready(self):
        """Returns True if Scheduler is ready to accept requests.

        This is to handle scheduler service startup when it has no volume hosts
        stats and will fail all the requests.
        """

        return self.host_manager.has_all_capabilities()

    def is_first_receive(self):
        """Returns True if Scheduler receives the capabilities at startup.

        This is to handle the problem of too long sleep time during scheduler
        service startup process.
        """

        return self.host_manager.first_receive_capabilities()

    def update_service_capabilities(self, service_name, host, capabilities,
                                    cluster_name, timestamp):
        """Process a capability update from a service node."""
        self.host_manager.update_service_capabilities(service_name,
                                                      host,
                                                      capabilities,
                                                      cluster_name,
                                                      timestamp)

    def notify_service_capabilities(self, service_name, backend,
                                    capabilities, timestamp):
        """Notify capability update from a service node."""
        self.host_manager.notify_service_capabilities(service_name,
                                                      backend,
                                                      capabilities,
                                                      timestamp)

    def host_passes_filters(self, context, backend, request_spec,
                            filter_properties):
        """Check if the specified backend passes the filters."""
        raise NotImplementedError(_("Must implement backend_passes_filters"))

    def find_retype_host(self, context, request_spec, filter_properties=None,
                         migration_policy='never'):
        """Find a backend that can accept the volume with its new type."""
        raise NotImplementedError(_("Must implement find_retype_backend"))

    # NOTE(geguileo): For backward compatibility with out of tree Schedulers
    # we don't change host_passes_filters or find_retype_host method names but
    # create an "alias" for them with the right name instead.
    backend_passes_filters = host_passes_filters
    find_retype_backend = find_retype_host

    def schedule(self, context, topic, method, *_args, **_kwargs):
        """Must override schedule method for scheduler to work."""
        raise NotImplementedError(_("Must implement a fallback schedule"))

    def schedule_create_volume(self, context, request_spec, filter_properties):
        """Must override schedule method for scheduler to work."""
        raise NotImplementedError(_("Must implement schedule_create_volume"))

    def schedule_create_group(self, context, group,
                              group_spec,
                              request_spec_list,
                              group_filter_properties,
                              filter_properties_list):
        """Must override schedule method for scheduler to work."""
        raise NotImplementedError(_(
            "Must implement schedule_create_group"))

    def get_pools(self, context, filters):
        """Must override schedule method for scheduler to work."""
        raise NotImplementedError(_(
            "Must implement schedule_get_pools"))

    def get_backup_host(self, volume, driver=None):
        """Must override schedule method for scheduler to work."""
        raise NotImplementedError(_(
            "Must implement get_backup_host"))



# Copyright (c) 2017 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from cinder.policies import attachments
from cinder.policies import backup_actions
from cinder.policies import backups
from cinder.policies import base
from cinder.policies import capabilities
from cinder.policies import clusters
from cinder.policies import default_types
from cinder.policies import group_actions
from cinder.policies import group_snapshot_actions
from cinder.policies import group_snapshots
from cinder.policies import group_types
from cinder.policies import groups
from cinder.policies import hosts
from cinder.policies import limits
from cinder.policies import manageable_snapshots
from cinder.policies import manageable_volumes
from cinder.policies import messages
from cinder.policies import qos_specs
from cinder.policies import quota_class
from cinder.policies import quotas
from cinder.policies import scheduler_stats
from cinder.policies import services
from cinder.policies import snapshot_actions
from cinder.policies import snapshot_metadata
from cinder.policies import snapshots
from cinder.policies import type_extra_specs
from cinder.policies import volume_access
from cinder.policies import volume_actions
from cinder.policies import volume_metadata
from cinder.policies import volume_transfer
from cinder.policies import volume_type
from cinder.policies import volumes
from cinder.policies import workers


def list_rules():
    return itertools.chain(
        base.list_rules(),
        attachments.list_rules(),
        messages.list_rules(),
        clusters.list_rules(),
        workers.list_rules(),
        snapshot_metadata.list_rules(),
        snapshots.list_rules(),
        snapshot_actions.list_rules(),
        manageable_snapshots.list_rules(),
        backups.list_rules(),
        backup_actions.list_rules(),
        groups.list_rules(),
        group_types.list_rules(),
        group_snapshots.list_rules(),
        group_snapshot_actions.list_rules(),
        group_actions.list_rules(),
        qos_specs.list_rules(),
        quota_class.list_rules(),
        quotas.list_rules(),
        capabilities.list_rules(),
        services.list_rules(),
        scheduler_stats.list_rules(),
        hosts.list_rules(),
        limits.list_rules(),
        manageable_volumes.list_rules(),
        volume_type.list_rules(),
        volume_access.list_rules(),
        volume_actions.list_rules(),
        volume_transfer.list_rules(),
        volume_metadata.list_rules(),
        type_extra_specs.list_rules(),
        volumes.list_rules(),
        default_types.list_rules(),
    )



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
from unittest import mock

import ddt

from cinder.api.contrib import volume_encryption_metadata
from cinder.api.contrib import volume_tenant_attribute
from cinder.api.v3 import volumes
from cinder import exception
from cinder.policies import volumes as volume_policies
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit import fake_constants
from cinder.tests.unit.policies import base
from cinder.tests.unit.policies import test_base
from cinder.tests.unit import utils as test_utils
from cinder.volume import api as volume_api


# TODO(yikun): The below policy test cases should be added:
# * HOST_ATTRIBUTE_POLICY
# * MIG_ATTRIBUTE_POLICY
class VolumePolicyTests(test_base.CinderPolicyTests):

    def test_admin_can_create_volume(self):
        admin_context = self.admin_context

        path = '/v3/%(project_id)s/volumes' % {
            'project_id': admin_context.project_id
        }
        body = {"volume": {"size": 1}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)

        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    def test_nonadmin_user_can_create_volume(self):
        user_context = self.user_context

        path = '/v3/%(project_id)s/volumes' % {
            'project_id': user_context.project_id
        }
        body = {"volume": {"size": 1}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)

        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    def test_admin_can_create_volume_from_image(self):
        admin_context = self.admin_context

        path = '/v3/%(project_id)s/volumes' % {
            'project_id': admin_context.project_id
        }
        body = {"volume": {"size": 1, "image_id": fake_constants.IMAGE_ID}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)

        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    def test_nonadmin_user_can_create_volume_from_image(self):
        user_context = self.user_context

        path = '/v3/%(project_id)s/volumes' % {
            'project_id': user_context.project_id
        }
        body = {"volume": {"size": 1, "image_id": fake_constants.IMAGE_ID}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)

        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get_volume')
    def test_admin_can_show_volumes(self, mock_volume):
        # Make sure administrators are authorized to list volumes
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(admin_context, path, 'GET')

        self.assertEqual(HTTPStatus.OK, response.status_int)
        self.assertEqual(response.json_body['volume']['id'], volume.id)

    @mock.patch.object(volume_api.API, 'get_volume')
    def test_owner_can_show_volumes(self, mock_volume):
        # Make sure owners are authorized to list their volumes
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(user_context, path, 'GET')

        self.assertEqual(HTTPStatus.OK, response.status_int)
        self.assertEqual(response.json_body['volume']['id'], volume.id)

    @mock.patch.object(volume_api.API, 'get_volume')
    def test_owner_cannot_show_volumes_for_others(self, mock_volume):
        # Make sure volumes are only exposed to their owners
        owner_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(owner_context)
        mock_volume.return_value = volume

        path = '/v3/%(project_id)s/volumes/%(volume_id)s' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(non_owner_context, path, 'GET')
        # NOTE(lbragstad): Technically, this user isn't supposed to see this
        # volume, because they didn't create it and it lives in a different
        # project. Does cinder return a 404 in cases like this? Or is a 403
        # expected?
        self.assertEqual(HTTPStatus.NOT_FOUND, response.status_int)

    def test_admin_can_get_all_volumes_detail(self):
        # Make sure administrators are authorized to list volumes
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        path = '/v3/%(project_id)s/volumes/detail' % {
            'project_id': admin_context.project_id
        }

        response = self._get_request_response(admin_context, path, 'GET')

        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_vol = response.json_body['volumes'][0]

        self.assertEqual(volume.id, res_vol['id'])

    def test_owner_can_get_all_volumes_detail(self):
        # Make sure owners are authorized to list volumes
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        path = '/v3/%(project_id)s/volumes/detail' % {
            'project_id': user_context.project_id
        }

        response = self._get_request_response(user_context, path, 'GET')

        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_vol = response.json_body['volumes'][0]

        self.assertEqual(volume.id, res_vol['id'])

    @mock.patch.object(volume_api.API, 'get')
    def test_admin_can_update_volumes(self, mock_volume):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"volume": {"name": "update_name"}}
        response = self._get_request_response(admin_context, path, 'PUT',
                                              body=body)
        self.assertEqual(HTTPStatus.OK, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_can_update_volumes(self, mock_volume):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"volume": {"name": "update_name"}}
        response = self._get_request_response(user_context, path, 'PUT',
                                              body=body)
        self.assertEqual(HTTPStatus.OK, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_update_volumes_for_others(self, mock_volume):
        owner_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(owner_context)
        mock_volume.return_value = volume

        path = '/v3/%(project_id)s/volumes/%(volume_id)s' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"volume": {"name": "update_name"}}
        response = self._get_request_response(non_owner_context, path, 'PUT',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_can_delete_volumes(self, mock_volume):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(user_context, path, 'DELETE')
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_admin_can_delete_volumes(self, mock_volume):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(admin_context, path, 'DELETE')
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_delete_volumes_for_others(self, mock_volume):
        owner_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(owner_context)
        mock_volume.return_value = volume

        path = '/v3/%(project_id)s/volumes/%(volume_id)s' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(non_owner_context, path,
                                              'DELETE')
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    @mock.patch.object(volume_api.API, 'get_volume')
    def test_admin_can_show_tenant_id_in_volume(self, mock_volume):
        # Make sure administrators are authorized to show tenant_id
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(admin_context, path, 'GET')

        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_vol = response.json_body['volume']
        self.assertEqual(admin_context.project_id,
                         res_vol['os-vol-tenant-attr:tenant_id'])

    @mock.patch.object(volume_api.API, 'get_volume')
    def test_owner_can_show_tenant_id_in_volume(self, mock_volume):
        # Make sure owners are authorized to show tenant_id in volume
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(user_context, path, 'GET')

        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_vol = response.json_body['volume']
        self.assertEqual(user_context.project_id,
                         res_vol['os-vol-tenant-attr:tenant_id'])

    def test_admin_can_show_tenant_id_in_volume_detail(self):
        # Make sure admins are authorized to show tenant_id in volume detail
        admin_context = self.admin_context

        self._create_fake_volume(admin_context)
        path = '/v3/%(project_id)s/volumes/detail' % {
            'project_id': admin_context.project_id
        }

        response = self._get_request_response(admin_context, path, 'GET')

        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_vol = response.json_body['volumes'][0]
        # Make sure owners are authorized to show tenant_id
        self.assertEqual(admin_context.project_id,
                         res_vol['os-vol-tenant-attr:tenant_id'])

    def test_owner_can_show_tenant_id_in_volume_detail(self):
        # Make sure owners are authorized to show tenant_id in volume detail
        user_context = self.user_context

        self._create_fake_volume(user_context)
        path = '/v3/%(project_id)s/volumes/detail' % {
            'project_id': user_context.project_id
        }

        response = self._get_request_response(user_context, path, 'GET')

        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_vol = response.json_body['volumes'][0]
        # Make sure owners are authorized to show tenant_id
        self.assertEqual(user_context.project_id,
                         res_vol['os-vol-tenant-attr:tenant_id'])

    def test_admin_can_create_metadata(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context, metadata={"k": "v"})
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"metadata": {"k1": "v1"}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.OK, response.status_int)

    def test_admin_can_get_metadata(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context, metadata={"k": "v"})
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(admin_context, path, 'GET')
        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_meta = response.json_body['metadata']
        self.assertIn('k', res_meta)
        self.assertEqual('v', res_meta['k'])

    def test_admin_can_update_metadata(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context, metadata={"k": "v"})
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"metadata": {"k": "v2"}}
        response = self._get_request_response(admin_context, path, 'PUT',
                                              body=body)
        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_meta = response.json_body['metadata']
        self.assertIn('k', res_meta)
        self.assertEqual('v2', res_meta['k'])

    def test_admin_can_delete_metadata(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context, metadata={"k": "v"})

        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata/%(key)s' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id,
            'key': 'k'
        }
        response = self._get_request_response(admin_context, path, 'DELETE')
        self.assertEqual(HTTPStatus.OK, response.status_int)

    def test_owner_can_create_metadata(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context, metadata={"k": "v"})
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"metadata": {"k1": "v1"}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.OK, response.status_int)

    def test_owner_can_get_metadata(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context, metadata={"k": "v"})
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(user_context, path, 'GET')
        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_meta = response.json_body['metadata']
        self.assertIn('k', res_meta)
        self.assertEqual('v', res_meta['k'])

    def test_owner_can_update_metadata(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context, metadata={"k": "v"})
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"metadata": {"k": "v2"}}
        response = self._get_request_response(user_context, path, 'PUT',
                                              body=body)
        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_meta = response.json_body['metadata']
        self.assertIn('k', res_meta)
        self.assertEqual('v2', res_meta['k'])

    def test_owner_can_delete_metadata(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context, metadata={"k": "v"})

        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata/%(key)s' % {
            'project_id': user_context.project_id, 'volume_id': volume.id,
            'key': 'k'
        }
        response = self._get_request_response(user_context, path, 'DELETE')
        self.assertEqual(HTTPStatus.OK, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_create_metadata_for_others(self, mock_volume):
        owner_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(owner_context, metadata={"k": "v"})
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"metadata": {"k1": "v1"}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_get_metadata_for_others(self, mock_volume):
        owner_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(owner_context, metadata={"k": "v"})
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(non_owner_context, path, 'GET')
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_update_metadata_for_others(self, mock_volume):
        owner_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(owner_context, metadata={"k": "v"})
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"metadata": {"k": "v2"}}
        response = self._get_request_response(non_owner_context, path, 'PUT',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_delete_metadata_for_others(self, mock_volume):
        owner_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(owner_context, metadata={"k": "v"})
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata/%(key)s' % {
            'project_id': non_owner_context.project_id,
            'volume_id': volume.id,
            'key': 'k'
        }
        response = self._get_request_response(non_owner_context, path,
                                              'DELETE')
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)


@ddt.ddt
class VolumesPolicyTest(base.BasePolicyTest):

    authorized_readers = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]
    unauthorized_readers = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]
    authorized_members = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]
    unauthorized_members = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    create_authorized_users = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
        # The other_* users are allowed because we don't have any check
        # mechanism in the code to validate this, these are validated on
        # the WSGI layer
        'other_project_member',
        'other_project_reader',
    ]

    create_unauthorized_users = [
        'system_member',
        'system_reader',
        'system_foo',
    ]

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.controller = volumes.VolumeController(mock.MagicMock())
        self.api_path = '/v3/%s/volumes' % (self.project_id)

    def _create_volume(self):
        vol_type = test_utils.create_volume_type(self.project_admin_context,
                                                 name='fake_vol_type',
                                                 testcase_instance=self)
        volume = test_utils.create_volume(self.project_member_context,
                                          volume_type_id=vol_type.id,
                                          testcase_instance=self)
        return volume

    @ddt.data(*base.all_users)
    def test_create_volume_policy(self, user_id):
        rule_name = volume_policies.CREATE_POLICY
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url)
        req.method = 'POST'
        body = {"volume": {"size": 1}}
        unauthorized_exceptions = []
        self.common_policy_check(user_id, self.create_authorized_users,
                                 self.create_unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.create, req,
                                 body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.api.v3.volumes.VolumeController._image_uuid_from_ref',
                return_value=fake_constants.IMAGE_ID)
    @mock.patch('cinder.api.v3.volumes.VolumeController._get_image_snapshot',
                return_value=None)
    @mock.patch('cinder.volume.flows.api.create_volume.'
                'ExtractVolumeRequestTask._get_image_metadata',
                return_value=None)
    def test_create_volume_from_image_policy(
            self, user_id, mock_image_from_ref, mock_image_snap,
            mock_img_meta):
        rule_name = volume_policies.CREATE_FROM_IMAGE_POLICY
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url)
        req.method = 'POST'
        body = {"volume": {"size": 1, "image_id": fake_constants.IMAGE_ID}}
        unauthorized_exceptions = []
        self.common_policy_check(user_id, self.create_authorized_users,
                                 self.create_unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.create, req,
                                 body=body)

    @ddt.data(*base.all_users)
    def test_create_multiattach_volume_policy(self, user_id):
        vol_type = test_utils.create_volume_type(
            self.project_admin_context, name='multiattach_type',
            extra_specs={'multiattach': '<is> True'})
        rule_name = volume_policies.MULTIATTACH_POLICY
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url)
        req.method = 'POST'
        body = {"volume": {"size": 1, "volume_type": vol_type.id}}

        # Relax the CREATE_POLICY in order to get past that check.
        self.policy.set_rules({volume_policies.CREATE_POLICY: ""},
                              overwrite=False)

        unauthorized_exceptions = []
        self.common_policy_check(user_id, self.create_authorized_users,
                                 self.create_unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.create, req,
                                 body=body)

    @ddt.data(*base.all_users)
    def test_get_volume_policy(self, user_id):
        volume = self._create_volume()
        rule_name = volume_policies.GET_POLICY
        url = '%s/%s' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url)
        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]
        self.common_policy_check(user_id,
                                 self.authorized_readers,
                                 self.unauthorized_readers,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.show, req,
                                 id=volume.id)

    @ddt.data(*base.all_users)
    def test_get_all_volumes_policy(self, user_id):
        self._create_volume()
        rule_name = volume_policies.GET_ALL_POLICY
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url)
        # Generally, any logged in user can list all volumes.
        authorized_users = [user_id]
        unauthorized_users = []
        # The exception is when deprecated rules are disabled, in which case
        # roles are enforced. Users without the 'reader' role should be
        # blocked.
        if self.enforce_new_defaults:
            context = self.create_context(user_id)
            if 'reader' not in context.roles:
                authorized_users = []
                unauthorized_users = [user_id]
        response = self.common_policy_check(user_id, authorized_users,
                                            unauthorized_users, [],
                                            rule_name,
                                            self.controller.index, req)
        # For some users, even if they're authorized, the list of volumes
        # will be empty if they are not in the volume's project.
        empty_response_users = [
            *self.unauthorized_readers,
            # legacy_admin and system_admin do not have a project_id, and
            # so the list of volumes returned will be empty.
            'legacy_admin',
            'system_admin',
        ]
        volumes = response['volumes'] if response else []
        volume_count = 0 if user_id in empty_response_users else 1
        self.assertEqual(volume_count, len(volumes))

    @ddt.data(*base.all_users)
    @mock.patch('cinder.db.volume_encryption_metadata_get')
    def test_get_volume_encryption_meta_policy(self, user_id,
                                               mock_encrypt_meta):
        encryption_key_id = fake_constants.ENCRYPTION_KEY_ID
        mock_encrypt_meta.return_value = (
            {'encryption_key_id': encryption_key_id})
        controller = (
            volume_encryption_metadata.VolumeEncryptionMetadataController())
        volume = self._create_volume()
        rule_name = volume_policies.ENCRYPTION_METADATA_POLICY
        url = '%s/%s/encryption' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url)
        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]
        resp = self.common_policy_check(
            user_id, self.authorized_readers,
            self.unauthorized_readers,
            unauthorized_exceptions,
            rule_name, controller.index, req,
            volume.id)
        if user_id in self.authorized_readers:
            self.assertEqual(encryption_key_id, resp['encryption_key_id'])

    @ddt.data(*base.all_users)
    def test_get_volume_tenant_attr_policy(self, user_id):
        controller = volume_tenant_attribute.VolumeTenantAttributeController()
        volume = self._create_volume()
        volume = volume.obj_to_primitive()['versioned_object.data']
        rule_name = volume_policies.TENANT_ATTRIBUTE_POLICY
        url = '%s/%s' % (self.api_path, volume['id'])
        req = fake_api.HTTPRequest.blank(url)
        req.get_db_volume = mock.MagicMock()
        req.get_db_volume.return_value = volume
        resp_obj = mock.MagicMock(obj={'volume': volume})
        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]
        self.assertNotIn('os-vol-tenant-attr:tenant_id', volume.keys())

        self.common_policy_check(
            user_id, self.authorized_readers,
            self.unauthorized_readers,
            unauthorized_exceptions,
            rule_name, controller.show, req,
            resp_obj, volume['id'], fatal=False)

        if user_id in self.authorized_readers:
            self.assertIn('os-vol-tenant-attr:tenant_id', volume.keys())

    @ddt.data(*base.all_users)
    def test_update_volume_policy(self, user_id):
        volume = self._create_volume()
        rule_name = volume_policies.UPDATE_POLICY
        url = '%s/%s' % (self.api_path, volume.id)
        body = {"volume": {"name": "update_name"}}
        req = fake_api.HTTPRequest.blank(url)
        req.method = 'PUT'
        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]
        self.common_policy_check(
            user_id, self.authorized_members,
            self.unauthorized_members,
            unauthorized_exceptions,
            rule_name, self.controller.update, req,
            id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_delete_volume_policy(self, user_id):
        volume = self._create_volume()
        rule_name = volume_policies.DELETE_POLICY
        url = '%s/%s' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url)
        req.method = 'DELETE'
        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]
        self.common_policy_check(
            user_id, self.authorized_members,
            self.unauthorized_members,
            unauthorized_exceptions,
            rule_name, self.controller.delete, req,
            id=volume.id)


class VolumesPolicySecureRbacTest(VolumesPolicyTest):
    create_authorized_users = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
        'other_project_member',
    ]

    create_unauthorized_users = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_reader',
        'project_foo',
        'project_reader',
    ]

    authorized_readers = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
    ]
    unauthorized_readers = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_members = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
    ]
    unauthorized_members = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)



#    (c) Copyright 2019 Brocade, a Broadcom Company
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""
Brocade south bound connector to communicate with switch using
REST over HTTP or HTTPS protocol.
"""

import json

from oslo_log import log as logging
from oslo_serialization import base64
import requests
import six

from cinder.i18n import _
from cinder.zonemanager.drivers.brocade import exception
from cinder.zonemanager.drivers.brocade import fc_zone_constants
from cinder.zonemanager.drivers.brocade import rest_constants

LOG = logging.getLogger(__name__)


class BrcdRestFCZoneClient(object):

    def __init__(self, ipaddress, username,
                 password, port, vfid, protocol):
        """Initializing the client with the parameters passed.

        :param ipaddress: IP Address of the device.
        :param username: User id to login.
        :param password: User password.
        :param port: Device Communication port
        :param vfid: Virtual Fabric ID.
        :param protocol: Communication Protocol.

        """
        self.sw_ip = ipaddress
        self.sw_user = username
        self.sw_pwd = password
        self.protocol = protocol
        self.vfid = vfid
        self.status_code = ''
        self.session = None
        self._login()

    def is_supported_firmware(self):
        is_supported_firmware = False
        fw_version = self._get_firmware_version()
        ver = fw_version.split(".")
        if len(ver[0]) > 1:
            major_ver = ver[0]
            ver[0] = major_ver[1]
        if len(ver[2]) > 1:
            patch_ver = ver[2]
            ver[2] = patch_ver[0]
        LOG.debug("Firmware version: %(version)s.", {'version': ver})
        if int(ver[0] + ver[1] + ver[2]) > 820:
            is_supported_firmware = True
        return is_supported_firmware

    def get_active_zone_set(self):
        active_zone_set, checksum = self._get_effective_zone_set()
        return active_zone_set

    def get_nameserver_info(self):
        return self._get_name_server()

    def add_zones(self, add_zone_map, activate, active_zone_set=None):
        self._add_zones(add_zone_map, activate)

    def update_zones(self, update_zone_map, activate, operation,
                     active_zone_set=None):
        self._update_zones(update_zone_map, activate, operation)

    def delete_zones(self, zone_names_to_delete, activate,
                     active_zone_set=None):
        self._delete_zones(zone_names_to_delete, activate)

    def cleanup(self):
        self._logout()

    def _login(self):
        if self.protocol == fc_zone_constants.REST_HTTPS:
            self.protocol = fc_zone_constants.HTTPS
        else:
            self.protocol = fc_zone_constants.HTTP
        if self.session is None:
            self.session = requests.Session()
            adapter = requests.adapters.HTTPAdapter(pool_connections=1,
                                                    pool_maxsize=1)
            self.session.mount(self.protocol + '://', adapter)
            credentials = base64.encode_as_text('%s:%s' % (self.sw_user,
                                                self.sw_pwd)).replace('\n', '')
            self.session.headers = {rest_constants.USER_AGENT:
                                    rest_constants.ZONE_DRIVER,
                                    rest_constants.ACCEPT: rest_constants.YANG,
                                    rest_constants.AUTHORIZATION:
                                        "Basic %s" % credentials}
        response = self.session.post(self._build_url(rest_constants.LOGIN))
        if response.status_code == 200:
            auth = response.headers.get('Authorization')
            LOG.info("REST login success, setting auth: %s", auth)
            self.session.headers = {rest_constants.USER_AGENT:
                                    rest_constants.ZONE_DRIVER,
                                    rest_constants.ACCEPT: rest_constants.YANG,
                                    rest_constants.CONTENT_TYPE:
                                        rest_constants.YANG,
                                    rest_constants.AUTHORIZATION: auth}
        else:
            msg = (_("REST login failed: %s")
                   % six.text_type(response.text))
            LOG.error(msg)
            raise exception.BrocadeZoningRestException(reason=msg)
        return response.status_code

    def _logout(self):
        response = self.session.post(self._build_url(rest_constants.LOGOUT))
        if response.status_code == 204:
            LOG.info("REST logout success")
        else:
            msg = (_("REST logout failed: %s")
                   % six.text_type(response.text))
            LOG.error(msg)
            raise exception.BrocadeZoningRestException(reason=msg)

    def _get_firmware_version(self):
        response = self.session.get(self._build_url(rest_constants.GET_SWITCH))
        firmware_version = ''
        if response.status_code == 200:
            data = response.json()
            json_response = data[rest_constants.RESPONSE]
            switch = json_response[rest_constants.SWITCH]
            firmware_version = switch[rest_constants.FIRMWARE_VERSION]
            LOG.info("REST firmware version: %s", firmware_version)
        else:
            msg = (_("REST get switch fw version failed: %s")
                   % six.text_type(response.text))
            LOG.error(msg)
            raise exception.BrocadeZoningRestException(reason=msg)
        return firmware_version

    def _get_name_server(self):
        port_names = []
        url = self._build_url(rest_constants.GET_NAMESERVER)
        response = self.session.get(url)
        if response.status_code == 200:
            data = response.json()
            json_response = data[rest_constants.RESPONSE]
            nsinfos = json_response[rest_constants.FC_NAME_SERVER]
            i = 0
            for nsinfo in nsinfos:
                port_names.append(nsinfos[i][rest_constants.PORT_NAME])
                i = i + 1
        else:
            msg = (_("REST get NS info failed: %s")
                   % six.text_type(response.text))
            LOG.error(msg)
            raise exception.BrocadeZoningRestException(reason=msg)
        return port_names

    def _get_effective_zone_set(self):
        active_zone_set = {}
        zones_map = {}
        url = self._build_url(rest_constants.GET_ACTIVE_ZONE_CFG)
        response = self.session.get(url)
        checksum = ''
        active_cfg_name = ''
        if response.status_code == 200:
            data = response.json()
            json_response = data[rest_constants.RESPONSE]
            effective_cfg = json_response[rest_constants.EFFECTIVE_CFG]
            checksum = effective_cfg[rest_constants.CHECKSUM]
            try:
                active_cfg_name = effective_cfg[rest_constants.CFG_NAME]
                zones = effective_cfg[rest_constants.ENABLED_ZONE]
                if type(zones) is list:
                    for i, zone in enumerate(zones):
                        zones_map.update({zones[i][rest_constants.ZONE_NAME]:
                                          zones[i][rest_constants.MEMBER_ENTRY]
                                                  [rest_constants.ENTRY_NAME]})
                else:
                    zones_map.update({zones[rest_constants.ZONE_NAME]:
                                      zones[rest_constants.MEMBER_ENTRY]
                                           [rest_constants.ENTRY_NAME]})
            except Exception:
                active_cfg_name = ''
            LOG.info("REST get effective zoneset success: "
                     "active cfg: %(cfg_name)s, checksum: %(chksum)s",
                     {'cfg_name': active_cfg_name, 'chksum': checksum})
        else:
            msg = (_("REST get effective zoneset failed: %s")
                   % six.text_type(response.text))
            LOG.error(msg)
            raise exception.BrocadeZoningRestException(reason=msg)
        active_zone_set = {"active_zone_config": active_cfg_name,
                           "zones": zones_map}
        return active_zone_set, checksum

    def _add_zones(self, add_zone_map, activate):
        active_zone_set, checksum = self._get_effective_zone_set()
        # if activate, get the zones already configured in the active cfg
        if activate:
            zones_in_active_cfg = active_zone_set.get("zones")
        # for each new zone, create a zone entry in defined zone db
        for zone_name, members in add_zone_map.items():
            if zone_name not in zones_in_active_cfg:
                body = {rest_constants.MEMBER_ENTRY:
                        {rest_constants.ENTRY_NAME:
                         add_zone_map.get(zone_name)}}
                json_str = json.dumps(body)
                url = self._build_url(rest_constants.POST_ZONE + zone_name)
                response = self.session.post(url, data=json_str)
                if response.status_code == 201:
                    LOG.info("REST create zone success: %s", zone_name)
                else:
                    msg = (_("REST create zone failed: %s")
                           % six.text_type(response.text))
                    LOG.error(msg)
                    raise exception.BrocadeZoningRestException(reason=msg)
        # update the cfg with the new zones
        active_cfg_name = active_zone_set.get("active_zone_config")
        active_zones = active_zone_set.get("zones")
        active_zone_names = active_zones.keys()
        active_zone_names.extend(add_zone_map.keys())
        body = {rest_constants.MEMBER_ZONE:
                {rest_constants.ZONE_NAME: active_zone_names}}
        json_str = json.dumps(body)
        if active_cfg_name == '':
            active_cfg_name = fc_zone_constants.CFG_NAME
            url = self._build_url(rest_constants.POST_CFG + active_cfg_name)
            response = self.session.post(url, data=json_str)
            if response.status_code == 201:
                LOG.info("REST cfg create success: %s", active_cfg_name)
                self._save_and_activate_cfg(checksum, activate,
                                            active_cfg_name)
            else:
                msg = (_("REST cfg create failed: %s")
                       % six.text_type(response.text))
                LOG.error(msg)
                raise exception.BrocadeZoningRestException(reason=msg)
        else:
            url = self._build_url(rest_constants.PATCH_CFG + active_cfg_name)
            response = self.session.patch(url, data=json_str)
            # if update successful, save the configuration changes
            if response.status_code == 204:
                LOG.info("REST cfg update success: %s", active_cfg_name)
                self._save_and_activate_cfg(checksum, activate,
                                            active_cfg_name)
            else:
                msg = (_("REST cfg update failed: %s")
                       % six.text_type(response.text))
                LOG.error(msg)
                raise exception.BrocadeZoningRestException(reason=msg)

    def _update_zones(self, update_zone_map, activate, operation):
        active_zone_set, checksum = self._get_effective_zone_set()
        active_cfg_name = active_zone_set.get("active_zone_config")
        active_zones = active_zone_set.get("zones")
        # for each zone, update the zone members in defined zone db
        for zone_name, members in update_zone_map.items():
            current_members = active_zones.get(zone_name)
            if operation == "ADD":
                new_members = set(members).difference(set(current_members))
                if new_members:
                    update_zone_map.update({zone_name: new_members})
            elif operation == "REMOVE":
                new_members = set(current_members).difference(set(members))
                if new_members:
                    update_zone_map.update({zone_name: new_members})
        # for each zone to be updated, make REST PATCH call to update
        for zone in update_zone_map.keys():
            body = {rest_constants.MEMBER_ENTRY:
                    {rest_constants.ENTRY_NAME: update_zone_map.get(zone)}}
            json_str = json.dumps(body)
            url = self._build_url(rest_constants.POST_ZONE + zone)
            response = self.session.patch(url, data=json_str)
            if response.status_code == 204:
                LOG.info("REST zone update success: %s", zone)
            else:
                msg = (_("REST zone update failed: %s")
                       % six.text_type(response.text))
                LOG.error(msg)
                raise exception.BrocadeZoningRestException(reason=msg)
        # save and activate the config changes
        self._save_and_activate_cfg(checksum, activate, active_cfg_name)

    def _delete_zones(self, zone_names_to_delete, activate):
        zone_names_to_delete = zone_names_to_delete.split(";")
        active_zone_set, checksum = self._get_effective_zone_set()
        # for each zone name, make REST DELETE call
        for zone in zone_names_to_delete:
            url = self._build_url(rest_constants.DELETE_ZONE + zone)
            response = self.session.delete(url)
            if response.status_code == 204:
                LOG.info("REST delete zone success: %s", zone)
            else:
                msg = (_("REST delete zone failed: %s")
                       % six.text_type(response.text))
                LOG.error(msg)
                raise exception.BrocadeZoningRestException(reason=msg)
        # update the cfg removing the deleted zones
        active_cfg_name = active_zone_set.get("active_zone_config")
        active_zones = active_zone_set.get("zones")
        active_zone_names = active_zones.keys()
        if len(active_zone_names) == len(zone_names_to_delete):
            # disable the cfg
            url = self._build_url(rest_constants.PATCH_CFG_DISABLE)
            body = {"checksum": checksum}
            json_str = json.dumps(body)
            response = self.session.patch(url, data=json_str)
            if response.status_code == 204:
                LOG.info("REST cfg disable success")
            else:
                msg = (_("REST cfg disable failed: %s")
                       % six.text_type(response.text))
                LOG.error(msg)
                raise exception.BrocadeZoningRestException(reason=msg)
            # delete the cfg
            url = self._build_url(rest_constants.DELETE_CFG + active_cfg_name)
            response = self.session.delete(url)
            if response.status_code == 204:
                LOG.info("REST cfg delete success: %s", active_cfg_name)
            else:
                msg = (_("REST cfg delete failed: %s")
                       % six.text_type(response.text))
                LOG.error(msg)
                raise exception.BrocadeZoningRestException(reason=msg)
            checksum = self._get_checksum()
            self._save_and_activate_cfg(checksum, False, active_cfg_name)
        else:
            # update the cfg by removing the deleted zones
            zone_names_in_cfg = list(set(active_zone_names)
                                     .difference(set(zone_names_to_delete)))
            body = {rest_constants.MEMBER_ZONE:
                    {rest_constants.ZONE_NAME: zone_names_in_cfg}}
            json_str = json.dumps(body)
            url = self._build_url(rest_constants.PATCH_CFG + active_cfg_name)
            response = self.session.patch(url, data=json_str)
            # if update successful, save the configuration changes
            if response.status_code == 204:
                LOG.info("REST cfg update success: %s", active_cfg_name)
                self._save_and_activate_cfg(checksum, activate,
                                            active_cfg_name)
            else:
                msg = (_("REST cfg update failed: %s")
                       % six.text_type(response.text))
                LOG.error(msg)
                raise exception.BrocadeZoningRestException(reason=msg)

    def _save_and_activate_cfg(self, checksum, activate, active_cfg_name):
        body = {"checksum": checksum}
        json_str = json.dumps(body)
        url = self._build_url(rest_constants.PATCH_CFG_SAVE)
        response = self.session.patch(url, data=json_str)
        if response.status_code == 204:
            LOG.info("REST cfg save success")
        else:
            msg = (_("REST cfg save failed: %s")
                   % six.text_type(response.text))
            LOG.error(msg)
            raise exception.BrocadeZoningRestException(reason=msg)
        # if activate=true, then enable the cfg changes to effective cfg
        if activate:
            checksum = self._get_checksum()
            body = {"checksum": checksum}
            json_str = json.dumps(body)
            url = self._build_url(rest_constants.PATCH_CFG_ENABLE
                                  + active_cfg_name)
            response = self.session.patch(url, data=json_str)
            if response.status_code == 204:
                LOG.info("REST cfg activate success: %s", active_cfg_name)
            else:
                msg = (_("REST cfg activate failed: %s")
                       % six.text_type(response.text))
                LOG.error(msg)
                raise exception.BrocadeZoningRestException(reason=msg)

    def _get_checksum(self):
        url = self._build_url(rest_constants.GET_CHECKSUM)
        response = self.session.get(url)
        checksum = ''
        if response.status_code == 200:
            data = response.json()
            json_response = data[rest_constants.RESPONSE]
            effective_cfg = json_response[rest_constants.EFFECTIVE_CFG]
            checksum = effective_cfg[rest_constants.CHECKSUM]
            LOG.info("REST get checksum success: %s", checksum)
        else:
            msg = (_("REST get checksum failed: %s")
                   % six.text_type(response.text))
            LOG.error(msg)
            raise exception.BrocadeZoningRestException(reason=msg)
        return checksum

    def _build_url(self, path):
        url = '%s://%s%s' % (self.protocol, self.sw_ip, path)
        if self.vfid is not None:
            url = '%s?vf-id=%s' % (url, self.vfid)
        return url



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Configuration options for Dell EMC PowerStore Cinder driver."""

from oslo_config import cfg

POWERSTORE_APPLIANCES = "powerstore_appliances"
POWERSTORE_PORTS = "powerstore_ports"
POWERSTORE_NVME = "powerstore_nvme"

POWERSTORE_OPTS = [
    cfg.ListOpt(POWERSTORE_APPLIANCES,
                default=[],
                help="Appliances names. Comma separated list of PowerStore "
                     "appliances names used to provision volumes.",
                deprecated_for_removal=True,
                deprecated_reason="Is not used anymore. "
                                  "PowerStore Load Balancer is used to "
                                  "provision volumes instead.",
                deprecated_since="Wallaby"),
    cfg.ListOpt(POWERSTORE_PORTS,
                default=[],
                help="Allowed ports. Comma separated list of PowerStore "
                     "iSCSI IPs or FC WWNs (ex. 58:cc:f0:98:49:22:07:02) "
                     "to be used. If option is not set all ports are allowed."
                ),
    cfg.BoolOpt(POWERSTORE_NVME,
                default=False,
                help="Connect PowerStore volumes using NVMe-OF.")
]



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from cinder import exception
from cinder.tests.unit import fake_group
from cinder.tests.unit import fake_group_snapshot
from cinder.tests.unit.volume.drivers.dell_emc import powerstore


class TestVolumeGroupSnapshotCreateDelete(powerstore.TestPowerStoreDriver):
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_chap_config")
    def setUp(self, mock_chap):
        super(TestVolumeGroupSnapshotCreateDelete, self).setUp()
        self.driver.check_for_setup_error()
        self.group = fake_group.fake_group_obj(
            self.context,
        )
        self.group_snapshot = fake_group_snapshot.fake_group_snapshot_obj(
            self.context
        )
        self.group_snapshot.group = self.group

    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.create_vg_snapshot")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_vg_id_by_name")
    @mock.patch("cinder.volume.volume_utils.is_group_a_cg_snapshot_type")
    def test_create_volume_group_snapshot(self,
                                          mock_is_cg,
                                          mock_get_id,
                                          mock_create):
        self.driver.create_group_snapshot(self.context,
                                          self.group_snapshot,
                                          [])

    @mock.patch("requests.request")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_vg_id_by_name")
    @mock.patch("cinder.volume.volume_utils.is_group_a_cg_snapshot_type")
    def test_create_volume_group_snapshot_bad_status(self,
                                                     mock_is_cg,
                                                     mock_get_id,
                                                     mock_create):
        mock_create.return_value = powerstore.MockResponse(rc=400)
        error = self.assertRaises(exception.VolumeBackendAPIException,
                                  self.driver.create_group_snapshot,
                                  self.context,
                                  self.group_snapshot,
                                  [])
        self.assertIn("Failed to create snapshot", error.msg)

    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.delete_volume_or_snapshot")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_vg_snapshot_id_by_name")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_vg_id_by_name")
    @mock.patch("cinder.volume.volume_utils.is_group_a_cg_snapshot_type")
    def test_delete_volume_group_snapshot(self,
                                          mock_is_cg,
                                          mock_get_group_id,
                                          mock_get_snapshot_id,
                                          mock_delete):
        self.driver.delete_group_snapshot(self.context,
                                          self.group_snapshot,
                                          [])

    @mock.patch("requests.request")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_vg_snapshot_id_by_name")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_vg_id_by_name")
    @mock.patch("cinder.volume.volume_utils.is_group_a_cg_snapshot_type")
    def test_delete_volume_group_snapshot_bad_status(self,
                                                     mock_is_cg,
                                                     mock_get_group_id,
                                                     mock_get_snapshot_id,
                                                     mock_delete):
        mock_delete.return_value = powerstore.MockResponse(rc=400)
        error = self.assertRaises(exception.VolumeBackendAPIException,
                                  self.driver.delete_group_snapshot,
                                  self.context,
                                  self.group_snapshot,
                                  [])
        self.assertIn("Failed to delete PowerStore volume group snapshot",
                      error.msg)



# Copyright 2018 Red Hat, Inc
# Copyright 2017 Rackspace Australia
# Copyright 2018 Michael Still and Aptira
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Helpers for ietadm related routines.
"""

from oslo_concurrency import processutils

import cinder.privsep


@cinder.privsep.sys_admin_pctxt.entrypoint
def new_target(name, tid):
    """Create new scsi target using specified parameters.

    If the target already exists, ietadm returns
    'Invalid argument' and error code '234'.
    This should be ignored for ensure export case.
    """
    processutils.execute('ietadm', '--op', 'new',
                         '--tid=%s' % tid,
                         '--params', 'Name=%s' % name,
                         check_exit_code=[0, 234])


@cinder.privsep.sys_admin_pctxt.entrypoint
def delete_target(tid):
    processutils.execute('ietadm', '--op', 'delete',
                         '--tid=%s' % tid)


@cinder.privsep.sys_admin_pctxt.entrypoint
def force_delete_target(tid, sid, cid):
    processutils.execute('ietadm', '--op', 'delete',
                         '--tid=%s' % tid,
                         '--sid=%s' % sid,
                         '--cid=%s' % cid)


@cinder.privsep.sys_admin_pctxt.entrypoint
def new_logicalunit(tid, lun, path, iotype):
    """Attach a new volume to scsi target as a logical unit.

    If a logical unit exists on the specified target lun,
    ietadm returns 'File exists' and error code '239'.
    This should be ignored for ensure export case.
    """

    processutils.execute('ietadm', '--op', 'new',
                         '--tid=%s' % tid,
                         '--lun=%d' % lun,
                         '--params',
                         'Path=%s,Type=%s' % (path, iotype),
                         check_exit_code=[0, 239])


@cinder.privsep.sys_admin_pctxt.entrypoint
def delete_logicalunit(tid, lun):
    processutils.execute('ietadm', '--op', 'delete',
                         '--tid=%s' % tid,
                         '--lun=%d' % lun)


@cinder.privsep.sys_admin_pctxt.entrypoint
def new_auth(tid, type, username, password):
    processutils.execute('ietadm', '--op', 'new',
                         '--tid=%s' % tid,
                         '--user',
                         '--params=%s=%s,Password=%s' % (type,
                                                         username,
                                                         password))
